@@ -5,6 +5,7 @@ import itertools
 import io
 import os
 import re
+import shutil
 
 from dulwich.errors import HangupException, GitProtocolError
 from dulwich.objects import Blob, Commit, Tag, Tree, parse_timezone
@@ -24,7 +25,9 @@ from mercurial import (
     error,
     phases,
     pycompat,
+    url,
     util as hgutil,
+    scmutil,
 )
 
 from . import _ssh
@@ -35,13 +38,13 @@ from . import util
 from .overlay import overlayrepo
 
 
-RE_GIT_AUTHOR = re.compile(b'^(.*?) ?\<(.*?)(?:\>(.*))?$')
+RE_GIT_AUTHOR = re.compile(br'^(.*?) ?\<(.*?)(?:\>(.*))?$')
 
-RE_GIT_SANITIZE_AUTHOR = re.compile(b'[<>\n]')
+RE_GIT_SANITIZE_AUTHOR = re.compile(br'[<>\n]')
 
-RE_GIT_AUTHOR_EXTRA = re.compile(b'^(.*?)\ ext:\((.*)\) <(.*)\>$')
+RE_GIT_AUTHOR_EXTRA = re.compile(br'^(.*?)\ ext:\((.*)\) <(.*)\>$')
 
-RE_GIT_EXTRA_KEY = re.compile(b'GIT([0-9]*)-(.*)')
+RE_GIT_EXTRA_KEY = re.compile(br'GIT([0-9]*)-(.*)')
 
 # Test for git:// and git+ssh:// URI.
 # Support several URL forms, including separating the
@@ -50,10 +53,10 @@ RE_GIT_URI = re.compile(
     br'^(?P<scheme>git([+]ssh)?://)(?P<host>.*?)(:(?P<port>\d+))?'
     br'(?P<sepr>[:/])(?P<path>.*)$')
 
-RE_NEWLINES = re.compile(b'[\r\n]')
-RE_GIT_PROGRESS = re.compile(b'\((\d+)/(\d+)\)')
+RE_NEWLINES = re.compile(br'[\r\n]')
+RE_GIT_PROGRESS = re.compile(br'\((\d+)/(\d+)\)')
 
-RE_AUTHOR_FILE = re.compile(b'\s*=\s*')
+RE_AUTHOR_FILE = re.compile(br'\s*=\s*')
 
 # mercurial.utils.dateutil functions were in mercurial.util in Mercurial < 4.6
 try:
@@ -108,6 +111,27 @@ class GitProgress(object):
             self.ui.note(msg + b'\n')
 
 
+def get_repo_and_gitdir(repo):
+    if repo.shared():
+        repo = compat.sharedreposource(repo)
+
+    if compat.config(repo.ui, b'bool', b'git', b'intree'):
+        gitdir = repo.wvfs.join(b'.git')
+    else:
+        gitdir = repo.vfs.join(b'git')
+
+    return repo, gitdir
+
+
+def has_gitrepo(repo):
+    if not hgutil.safehasattr(repo, b'vfs'):
+        return False
+
+    repo, gitdir = get_repo_and_gitdir(repo)
+
+    return os.path.isdir(gitdir)
+
+
 class GitHandler(object):
     map_file = b'git-mapfile'
     remote_refs_file = b'git-remote-refs'
@@ -115,21 +139,11 @@ class GitHandler(object):
 
     def __init__(self, dest_repo, ui):
         self.repo = dest_repo
+        self.store_repo, self.gitdir = get_repo_and_gitdir(self.repo)
         self.ui = ui
-        self.vfs = self.repo.vfs
-
-        # Mercurial >= 3.3:  repo.shared()
-        if dest_repo.sharedpath != dest_repo.path:
-            self.vfs = compat.hgvfs(dest_repo.sharedpath)
-        if compat.config(ui, b'bool', b'git', b'intree'):
-            self.gitdir = self.repo.wvfs.join(b'.git')
-        else:
-            self.gitdir = self.vfs.join(b'git')
 
         self.init_author_file()
 
-        self.paths = ui.configitems(b'paths')
-
         self.branch_bookmark_suffix = compat.config(
             ui, b'string', b'git', b'branch_bookmark_suffix')
 
@@ -138,6 +152,15 @@ class GitHandler(object):
         self.load_tags()
         self._remote_refs = None
 
+        # the HTTP authentication realm -- this specifies that we've
+        # tried an unauthenticated request, gotten a realm, and are now
+        # ready to prompt the user, if necessary
+        self._http_auth_realm = None
+
+    @property
+    def vfs(self):
+        return self.store_repo.vfs
+
     @property
     def _map_git(self):
         if self._map_git_real is None:
@@ -177,7 +200,7 @@ class GitHandler(object):
         self.author_map = {}
         authors_path = compat.config(self.ui, b'string', b'git', b'authors')
         if authors_path:
-            with open(self.repo.wvfs.join(authors_path)) as f:
+            with open(self.repo.wvfs.join(authors_path), 'rb') as f:
                 for line in f:
                     line = line.strip()
                     if not line or line.startswith(b'#'):
@@ -215,20 +238,12 @@ class GitHandler(object):
         self._map_hg_real = map_hg_real
 
     def save_map(self, map_file):
-        wlock = self.repo.wlock()
-        try:
-            file = self.vfs(map_file, b'w+', atomictemp=True)
+        with self.store_repo.wlock():
             map_hg = self._map_hg
-            buf = io.BytesIO()
-            bwrite = buf.write
-            for hgsha, gitsha in compat.iteritems(map_hg):
-                bwrite(b"%s %s\n" % (gitsha, hgsha))
-            file.write(buf.getvalue())
-            buf.close()
-            # If this complains, atomictempfile no longer has close
-            file.close()
-        finally:
-            wlock.release()
+            with self.vfs(map_file, b'wb+', atomictemp=True) as buf:
+                bwrite = buf.write
+                for hgsha, gitsha in compat.iteritems(map_hg):
+                    bwrite(b"%s %s\n" % (gitsha, hgsha))
 
     def load_tags(self):
         self.tags = {}
@@ -238,27 +253,24 @@ class GitHandler(object):
                 self.tags[name] = sha
 
     def save_tags(self):
-        file = self.vfs(self.tags_file, b'w+', atomictemp=True)
-        for name, sha in sorted(compat.iteritems(self.tags)):
-            if not self.repo.tagtype(name) == b'global':
-                file.write(b"%s %s\n" % (sha, name))
-        # If this complains, atomictempfile no longer has close
-        file.close()
+        with self.repo.wlock(), self.store_repo.wlock():
+            with self.vfs(self.tags_file, b'w+', atomictemp=True) as fp:
+                for name, sha in sorted(compat.iteritems(self.tags)):
+                    if not self.repo.tagtype(name) == b'global':
+                        fp.write(b"%s %s\n" % (sha, name))
 
     def load_remote_refs(self):
         self._remote_refs = {}
         refdir = os.path.join(self.gitdir, b'refs', b'remotes')
 
-        paths = self.paths
-        # if paths are set, we should still check 'default'
-        if not paths:
-            paths = ((b'default', None),)
+        # if no paths are set, we should still check 'default'
+        pathnames = list(self.ui.paths) or [b'default']
 
         # we avoid using dulwich's refs method because it is incredibly slow;
         # on a repo with a few hundred branches and a few thousand tags,
         # dulwich took about 200ms
-        for p in paths:
-            remotedir = os.path.join(refdir, p[0])
+        for pathname in pathnames:
+            remotedir = os.path.join(refdir, pathname)
             for root, dirs, files in os.walk(remotedir):
                 for f in files:
                     try:
@@ -275,13 +287,15 @@ class GitHandler(object):
     def import_commits(self, remote_name):
         refs = self.git.refs.as_dict()
         filteredrefs = self.filter_min_date(refs)
-        self.import_git_objects(remote_name, filteredrefs)
-        self.update_hg_bookmarks(refs)
-        self.save_map(self.map_file)
+        try:
+            self.import_git_objects(remote_name, filteredrefs)
+            self.update_hg_bookmarks(refs)
+        finally:
+            self.save_map(self.map_file)
 
     def fetch(self, remote, heads):
         result = self.fetch_pack(remote, heads)
-        remote_name = self.remote_name(remote)
+        remote_name = self.remote_name(remote, False)
 
         # if remote returns a symref for HEAD, then let's store that
         rhead = None
@@ -376,7 +390,6 @@ class GitHandler(object):
 
     def get_refs(self, remote):
         self.export_commits()
-        client, path = self.get_transport_and_path(remote)
         old_refs = {}
         new_refs = {}
 
@@ -387,16 +400,34 @@ class GitHandler(object):
             return refs  # always return the same refs to make the send a no-op
 
         try:
-            client.send_pack(path, changed, lambda have, want: [])
+            self._call_client(remote, 'send_pack', changed, lambda have, want: [])
 
             changed_refs = [ref for ref, sha in compat.iteritems(new_refs)
                             if sha != old_refs.get(ref)]
-            new = [bin(self.map_hg_get(new_refs[ref])) for ref in changed_refs]
+
+            new = [
+                bin(sha) for sha in map(self.map_hg_get, changed_refs) if sha
+            ]
             old = {}
-            for r in old_refs:
-                old_ref = self.map_hg_get(old_refs[r])
-                if old_ref:
-                    old[bin(old_ref)] = 1
+
+            for ref, sha in compat.iteritems(old_refs):
+                try:
+                    gittag = self.git.get_object(sha)
+                except KeyError:
+                    gittag = None
+
+                # make sure we don't accidentally dereference and lose
+                # annotated tags
+                if isinstance(gittag, Tag):
+                    sha = gittag.object[1]
+
+                old_target = self.map_hg_get(sha)
+
+                if old_target:
+                    self.ui.debug(b'unchanged ref %s: %s\n' % (ref, old_target))
+                    old[bin(old_target)] = 1
+                else:
+                    self.ui.debug(b'changed ref %s\n' % (ref))
 
             return old, new
         except (HangupException, GitProtocolError) as e:
@@ -406,7 +437,12 @@ class GitHandler(object):
     def push(self, remote, revs, force):
         self.export_commits()
         old_refs, new_refs = self.upload_pack(remote, revs, force)
-        remote_name = self.remote_name(remote)
+        remote_name = self.remote_name(remote, True)
+
+        if not isinstance(new_refs, dict):
+            # dulwich 0.20.6 changed the API and deprectated treating
+            # the result as a dictionary
+            new_refs = new_refs.refs
 
         if remote_name and new_refs:
             for ref, new_sha in sorted(compat.iteritems(new_refs)):
@@ -441,15 +477,13 @@ class GitHandler(object):
 
     def clear(self):
         mapfile = self.vfs.join(self.map_file)
+        tagsfile = self.vfs.join(self.tags_file)
         if os.path.exists(self.gitdir):
-            for root, dirs, files in os.walk(self.gitdir, topdown=False):
-                for name in files:
-                    os.remove(os.path.join(root, name))
-                for name in dirs:
-                    os.rmdir(os.path.join(root, name))
-            os.rmdir(self.gitdir)
+            shutil.rmtree(self.gitdir)
         if os.path.exists(mapfile):
             os.remove(mapfile)
+        if os.path.exists(tagsfile):
+            os.remove(tagsfile)
 
     # incoming support
     def getremotechanges(self, remote, revs):
@@ -621,7 +655,8 @@ class GitHandler(object):
         self.swap_out_encoding(oldenc)
         return commit.id
 
-    def get_valid_git_username_email(self, name):
+    @staticmethod
+    def get_valid_git_username_email(name):
         r"""Sanitize usernames and emails to fit git's restrictions.
 
         The following is taken from the man page of git's fast-import
@@ -652,13 +687,7 @@ class GitHandler(object):
 
         TESTS:
 
-        >>> from collections import namedtuple
-        >>> from mercurial.ui import ui
-        >>> mockrepo = namedtuple('localrepo', ['vfs', 'sharedpath', 'path'])
-        >>> mockrepo.vfs = namedtuple('vfs', ['join'])(lambda x: x)
-        >>> mockrepo.sharedpath = b''
-        >>> mockrepo.path = b''
-        >>> g = GitHandler(mockrepo, ui()).get_valid_git_username_email
+        >>> g = GitHandler.get_valid_git_username_email
         >>> g(b'John Doe')
         'John Doe'
         >>> g(b'john@doe.com')
@@ -828,6 +857,7 @@ class GitHandler(object):
 
     def import_git_commit(self, commit):
         self.ui.debug(_(b"importing: %s\n") % commit.id)
+        unfiltered = self.repo.unfiltered()
 
         detect_renames = False
         (strip_message, hg_renames,
@@ -846,7 +876,7 @@ class GitHandler(object):
         gparents = pycompat.maplist(self.map_hg_get, commit.parents)
 
         for parent in gparents:
-            if parent not in self.repo:
+            if parent not in unfiltered:
                 raise error.Abort(_(b'you appear to have run strip - '
                                     b'please run hg git-cleanup'))
 
@@ -867,7 +897,7 @@ class GitHandler(object):
             self.git_file_readlines(git_commit_tree, b'.hgsubstate'))
         parentsubdata = b''
         if gparents:
-            p1ctx = self.repo[gparents[0]]
+            p1ctx = unfiltered[gparents[0]]
             if b'.hgsubstate' in p1ctx:
                 parentsubdata = p1ctx.filectx(b'.hgsubstate').data()
                 parentsubdata = parentsubdata.splitlines()
@@ -914,7 +944,7 @@ class GitHandler(object):
         except UnicodeDecodeError:
             text = self.decode_guess(text, commit.encoding)
 
-        text = b'\n'.join([l.rstrip() for l in text.splitlines()]).strip(b'\n')
+        text = b'\n'.join(l.rstrip() for l in text.splitlines()).strip(b'\n')
         if text + b'\n' != origtext:
             extra[b'message'] = create_delta(text + b'\n', origtext)
 
@@ -951,8 +981,8 @@ class GitHandler(object):
             # begin with).
             if p2 == nullid:
                 return []
-            manifest1 = self.repo[p1].manifest()
-            manifest2 = self.repo[p2].manifest()
+            manifest1 = unfiltered[p1].manifest()
+            manifest2 = unfiltered[p2].manifest()
             return [path for path, node1 in compat.iteritems(manifest1) if path not
                     in files and manifest2.get(path, node1) != node1]
 
@@ -977,7 +1007,7 @@ class GitHandler(object):
                     e = self.convert_git_int_mode(mode)
             else:
                 # it's a converged file
-                fc = context.filectx(self.repo, f, changeid=memctx.p1().rev())
+                fc = context.filectx(unfiltered, f, changeid=memctx.p1().rev())
                 data = fc.data()
                 e = fc.flags()
                 copied_path = None
@@ -985,7 +1015,7 @@ class GitHandler(object):
                 if copied:
                     copied_path = copied[0]
 
-            return compat.memfilectx(self.repo, memctx, f, data,
+            return compat.memfilectx(unfiltered, memctx, f, data,
                                      islink=b'l' in e,
                                      isexec=b'x' in e,
                                      copysource=copied_path)
@@ -996,12 +1026,12 @@ class GitHandler(object):
         if len(gparents) > 1:
             # merge, possibly octopus
             def commit_octopus(p1, p2):
-                ctx = context.memctx(self.repo, (p1, p2), text, list(files) +
+                ctx = context.memctx(unfiltered, (p1, p2), text, list(files) +
                                      findconvergedfiles(p1, p2), getfilectx,
                                      author, date, {b'hg-git': b'octopus'})
                 # See comment below about setting substate to None.
                 ctx.substate = None
-                return hex(self.repo.commitctx(ctx))
+                return hex(unfiltered.commitctx(ctx))
 
             octopus = len(gparents) > 2
             p2 = gparents.pop()
@@ -1034,7 +1064,7 @@ class GitHandler(object):
         if octopus:
             extra[b'hg-git'] = b'octopus-done'
 
-        ctx = context.memctx(self.repo, (p1, p2), text,
+        ctx = context.memctx(unfiltered, (p1, p2), text,
                              list(files) + findconvergedfiles(p1, p2),
                              getfilectx, author, date, extra)
         # Starting Mercurial commit d2743be1bb06, memctx imports from
@@ -1045,7 +1075,7 @@ class GitHandler(object):
         # that won't work. Forcibly set the substate to None so that there's no
         # attempt to read subrepos.
         ctx.substate = None
-        node = self.repo.commitctx(ctx)
+        node = unfiltered.commitctx(ctx)
 
         self.swap_out_encoding(oldenc)
 
@@ -1056,7 +1086,6 @@ class GitHandler(object):
     # PACK UPLOADING AND FETCHING
 
     def upload_pack(self, remote, revs, force):
-        client, path = self.get_transport_and_path(remote)
         old_refs = {}
         change_totals = {}
 
@@ -1112,8 +1141,9 @@ class GitHandler(object):
                 self.ui.status(_(b"remote: %s\n") % line)
 
         try:
-            new_refs = client.send_pack(path, changed, genpack,
-                                        progress=callback)
+            new_refs = self._call_client(remote, 'send_pack', changed, genpack,
+                                         progress=callback)
+
             if len(change_totals) > 0:
                 self.ui.status(_(b"added %d commits with %d trees"
                                  b" and %d blobs\n") %
@@ -1137,7 +1167,7 @@ class GitHandler(object):
         # the future.)
         if not refs or next(iter(refs.keys())) == b'capabilities^{}':
             if not exportable:
-                tip = self.repo.lookup(b'tip')
+                tip = self.repo.filtered(b'served').lookup(b'tip')
                 if tip != nullid:
                     if b'capabilities^{}' in new_refs:
                         del new_refs[b'capabilities^{}']
@@ -1180,7 +1210,11 @@ class GitHandler(object):
 
             for ref in rev_refs:
                 if ref not in refs:
-                    new_refs[ref] = self.map_git_get(ctx.hex())
+                    gitobj = self.git.get_object(self.git.refs[ref])
+                    if isinstance(gitobj, Tag):
+                        new_refs[ref] = gitobj.id
+                    else:
+                        new_refs[ref] = self.map_git_get(ctx.hex())
                 elif new_refs[ref] in self._map_git:
                     rctx = unfiltered[self.map_hg_get(new_refs[ref])]
                     if rctx.ancestor(ctx) == rctx or force:
@@ -1199,8 +1233,6 @@ class GitHandler(object):
         return new_refs
 
     def fetch_pack(self, remote_name, heads=None):
-        localclient, path = self.get_transport_and_path(remote_name)
-
         # The dulwich default walk only checks refs/heads/. We also want to
         # consider remotes when doing discovery, so we build our own list. We
         # can't just do 'refs/' here because the tag class doesn't have a
@@ -1220,8 +1252,9 @@ class GitHandler(object):
             progress = GitProgress(self.ui)
             f = io.BytesIO()
 
-            ret = localclient.fetch_pack(path, determine_wants, graphwalker,
-                                         f.write, progress.progress)
+            ret = self._call_client(remote_name, 'fetch_pack', determine_wants,
+                                    graphwalker, f.write, progress.progress)
+
             if(f.tell() != 0):
                 f.seek(0)
                 self.git.object_store.add_thin_pack(f.read, None)
@@ -1238,6 +1271,57 @@ class GitHandler(object):
             raise error.Abort(_(b"git remote error: ")
                               + pycompat.sysbytes(str(e)))
 
+    def _call_client(self, remote_name, method, *args, **kwargs):
+        clientobj, path = self._get_transport_and_path(remote_name)
+
+        func = getattr(clientobj, method)
+
+        # dulwich 0.19, used in python 2.7, does not offer a specific
+        # exception class
+        HTTPUnauthorized = getattr(
+            client, 'HTTPUnauthorized', type('<dummy>', (Exception,), {}),
+        )
+
+        try:
+            return func(path, *args, **kwargs)
+        except (HTTPUnauthorized, GitProtocolError) as e:
+            self.ui.traceback()
+
+            if isinstance(e, HTTPUnauthorized):
+                # this is a fallback just in case the header isn't
+                # specified
+                self._http_auth_realm = 'Git'
+                if e.www_authenticate:
+                    m = re.search(r'realm="([^"]*)"', e.www_authenticate)
+                    if m:
+                        self._http_auth_realm = m.group(1)
+
+            elif 'unexpected http resp 407' in e.args[0]:
+                raise error.Abort(
+                    b'HTTP proxy requires authentication',
+                )
+            # dulwich 0.19
+            elif 'unexpected http resp 401' in e.args[0]:
+                self._http_auth_realm = 'Git'
+            else:
+                raise
+
+            clientobj, path = self._get_transport_and_path(remote_name)
+            func = getattr(clientobj, method)
+
+            try:
+                return func(path, *args, **kwargs)
+            except HTTPUnauthorized:
+                raise error.Abort(_(b'authorization failed'))
+            except GitProtocolError as e:
+                # python 2.7
+                if 'unexpected http resp 401' in e.args[0]:
+                    raise error.Abort(_(b'authorization failed'))
+                else:
+                    raise
+
+
+
     # REFERENCES HANDLING
 
     def filter_refs(self, refs, heads):
@@ -1316,19 +1400,38 @@ class GitHandler(object):
             if self.repo.tagtype(tag) in (b'global', b'git'):
                 tag = tag.replace(b' ', b'_')
                 target = self.map_git_get(hex(sha))
-                if target is not None:
-                    tag_refname = b'refs/tags/' + tag
-                    if(check_ref_format(tag_refname)):
-                        self.git.refs[tag_refname] = target
-                        self.tags[tag] = hex(sha)
-                    else:
-                        self.repo.ui.warn(b'Skipping export of tag %s because '
-                                          b'it has invalid name as a git '
-                                          b'refname.\n' % tag)
-                else:
-                    self.repo.ui.warn(
-                        b'Skipping export of tag %s because it '
-                        b'has no matching git revision.\n' % tag)
+
+                if target is None:
+                    self.repo.ui.warn(b"warning: not exporting tag '%s' "
+                                      b"due to missing git "
+                                      b"revision\n" % tag)
+                    continue
+
+                tag_refname = b'refs/tags/' + tag
+
+                if not check_ref_format(tag_refname):
+                    self.repo.ui.warn(b"warning: not exporting tag '%s' "
+                                      b"due to invalid name\n" % tag)
+                    continue
+
+                # check whether the tag already exists and is
+                # annotated
+                if tag_refname in self.git.refs:
+                    gittarget = self.git.refs[tag_refname]
+                    gittag = self.git.get_object(gittarget)
+                    if isinstance(gittag, Tag):
+                        if gittag.object[1] != target:
+                            self.repo.ui.warn(
+                                b"warning: not overwriting annotated "
+                                b"tag '%s'\n" % tag
+                            )
+
+                        # never overwrite annotated tags, otherwise
+                        # it'd happen on every pull
+                        target = gittarget
+
+                self.git.refs[tag_refname] = target
+                self.tags[tag] = hex(sha)
 
     def _filter_for_bookmarks(self, bms):
         if not self.branch_bookmark_suffix:
@@ -1358,7 +1461,13 @@ class GitHandler(object):
 
         bms = self.repo._bookmarks
         for filtered_bm, bm in self._filter_for_bookmarks(bms):
-            res[hex(bms[bm])].heads.add(b'refs/heads/' + filtered_bm)
+            ref_name = b'refs/heads/' + filtered_bm
+            if check_ref_format(ref_name):
+                res[hex(bms[bm])].heads.add(ref_name)
+            else:
+                self.repo.ui.warn(b"warning: not exporting bookmark '%s' "
+                                  b"due to invalid name\n" % bm)
+
         for tag, sha in compat.iteritems(self.tags):
             res[sha].tags.add(b'refs/tags/' + tag)
         return res
@@ -1371,7 +1480,7 @@ class GitHandler(object):
             ref_name = k
             parts = k.split(b'/')
             if parts[0] == b'refs' and parts[1] == b'tags':
-                ref_name = b"/".join([v for v in parts[2:]])
+                ref_name = b"/".join(v for v in parts[2:])
                 # refs contains all the refs in the server, not just
                 # the ones we are pulling
                 if refs[k] not in self.git.object_store:
@@ -1395,12 +1504,34 @@ class GitHandler(object):
                                 self.tags[ref_name] = sha
         self.save_tags()
 
+    def add_tag(self, target, *tags):
+        for tag in tags:
+            scmutil.checknewlabel(self.repo, tag, b'tag')
+
+            # -f/--force is deliberately unimplemented and unmentioned
+            # as its git semantics are quite confusing
+            if compat.isrevsymbol(self.repo, tag):
+                raise error.Abort(b"the name '%s' already exists" % tag)
+
+            if check_ref_format(b'refs/tags/' + tag):
+                self.ui.debug(b'adding git tag %s\n' % tag)
+                self.tags[tag] = target
+            else:
+                raise error.Abort(b"the name '%s' is not a valid git "
+                                  b"tag" % tag)
+
+        self.export_commits()
+        self.save_tags()
+
     def update_hg_bookmarks(self, refs):
         try:
             bms = self.repo._bookmarks
 
-            heads = dict([(ref[11:], refs[ref]) for ref in refs
-                          if ref.startswith(b'refs/heads/')])
+            heads = {
+                ref[11:]: refs[ref]
+                for ref in refs
+                if ref.startswith(b'refs/heads/')
+            }
 
             suffix = self.branch_bookmark_suffix or b''
             changes = []
@@ -1624,13 +1755,18 @@ class GitHandler(object):
             return content.splitlines()
         return []
 
-    def remote_name(self, remote):
-        names = [name for name, path in self.paths if path == remote]
-        if names:
-            return names[0]
+    def remote_name(self, remote, push):
+        for name, paths in compat.iteritems(self.ui.paths):
+            # paths became lists in mercurial 5.9
+            if not isinstance(paths, list):
+                paths = [paths]
+
+            for path in paths:
+                if push and path.pushloc == remote or path.loc == remote:
+                    return name
 
     def audit_hg_path(self, path):
-        if b'.hg' in path.split(pycompat.ossep):
+        if b'.hg' in path.split(b'/'):
             if compat.config(self.ui, b'bool', b'git', b'blockdothg'):
                 raise error.Abort(
                     (b"Refusing to import problematic path '%s'" % path),
@@ -1665,31 +1801,32 @@ class GitHandler(object):
         except UnicodeDecodeError:
             return string.decode('ascii', 'replace').encode('utf-8')
 
-    def get_transport_and_path(self, uri):
+    def _get_transport_and_path(self, uri):
         """Method that sets up the transport (either ssh or http(s))
 
         Tests:
 
-        >>> from collections import namedtuple
         >>> from dulwich.client import HttpGitClient, SSHGitClient
-        >>> from mercurial.ui import ui
-        >>> mockrepo = namedtuple('localrepo', ['vfs', 'sharedpath', 'path'])
-        >>> mockrepo.vfs = namedtuple('vfs', ['join'])(lambda x: x)
-        >>> mockrepo.sharedpath = b''
-        >>> mockrepo.path = b''
-        >>> g = GitHandler(mockrepo, ui())
-        >>> tp = g.get_transport_and_path
+        >>> from mercurial import ui
+        >>> class SubHandler(GitHandler):
+        ...    def __init__(self):
+        ...         self.ui = ui.ui()
+        ...         self._http_auth_realm = None
+        >>> tp = SubHandler()._get_transport_and_path
         >>> client, url = tp(b'http://fqdn.com/test.git')
         >>> print(isinstance(client, HttpGitClient))
         True
         >>> print(url.decode())
         http://fqdn.com/test.git
+        >>> client, url = tp(b'c:/path/to/repo.git')
+        >>> print(isinstance(client, SSHGitClient))
+        False
         >>> client, url = tp(b'git@fqdn.com:user/repo.git')
         >>> print(isinstance(client, SSHGitClient))
         True
         >>> print(url.decode())
         user/repo.git
-        >>> print(client.host.decode())
+        >>> print(client.host)
         git@fqdn.com
         """
         # pass hg's ui.ssh config to dulwich
@@ -1706,7 +1843,7 @@ class GitHandler(object):
             host, port, sepr = res['host'], res['port'], res['sepr']
             transport = client.TCPGitClient
             if b'ssh' in res['scheme']:
-                util.checksafessh(host)
+                util.checksafessh(pycompat.bytesurl(host))
                 transport = client.SSHGitClient
             path = res['path']
             if sepr == b'/' and not path.startswith(b'~'):
@@ -1718,7 +1855,7 @@ class GitHandler(object):
             if port:
                 client.port = port
 
-            return transport(host, port=port), path
+            return transport(pycompat.strurl(host), port=port), path
 
         if uri.startswith(b'git+http://') or uri.startswith(b'git+https://'):
             uri = uri[4:]
@@ -1727,22 +1864,50 @@ class GitHandler(object):
             ua = b'git/20x6 (hg-git ; uses dulwich and hg ; like git-core)'
             config = dul_config.ConfigDict()
             config.set(b'http', b'useragent', ua)
+
+            proxy = compat.config(self.ui, b'string', b'http_proxy', b'host')
+            if proxy:
+                config.set(b'http', b'proxy', b'http://' + proxy)
+
+                if compat.config(self.ui, b'string', b'http_proxy', b'passwd'):
+                    self.ui.warn(
+                        b"warning: proxy authentication is unsupported\n",
+                    )
+
             if pycompat.ispy3:
                 # urllib3.util.url._encode_invalid_chars() converts the path
                 # back to bytes using the utf-8 codec
                 str_uri = uri.decode('utf-8')
             else:
                 str_uri = uri
-            url = hgutil.url(uri)
-            username = url.user
-            password = url.passwd
-            if pycompat.ispy3:
-                # urllib3.util.request.make_headers() converts them back to
-                # bytes using the latin-1 codec
-                if username is not None:
-                    username = username.decode('latin-1')
-                if password is not None:
-                    password = password.decode('latin-1')
+
+            pwmgr = url.passwordmgr(self.ui, self.ui.httppasswordmgrdb)
+
+            # not available in dulwich 0.19, used on Python 2.7
+            if hasattr(client, 'get_credentials_from_store'):
+                urlobj = compat.url(uri)
+                auth = client.get_credentials_from_store(
+                    urlobj.scheme,
+                    urlobj.host,
+                    urlobj.user,
+                )
+            else:
+                auth = None
+
+            if self._http_auth_realm:
+                # since we've tried an unauthenticated request, and
+                # obtain a realm, we can do a "full" search, including
+                # a prompt
+                username, password = pwmgr.find_user_password(
+                    self._http_auth_realm, str_uri,
+                )
+            elif auth is not None:
+                username, password = auth
+                username = username.decode('utf-8')
+                password = password.decode('utf-8')
+            else:
+                username, password = pwmgr.find_stored_password(str_uri)
+
             return (
                 client.HttpGitClient(
                     str_uri,
@@ -1753,5 +1918,8 @@ class GitHandler(object):
                 uri,
             )
 
+        if uri.startswith(b'file://'):
+            return client.LocalGitClient(), compat.url(uri).path
+
         # if its not git or git+ssh, try a local url..
         return client.SubprocessGitClient(), uri
